# coding=utf-8
# Copyright 2013 Gridward LLC
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from uuid import UUID
from collections import defaultdict
import os
import logging
import datetime as dt

from lxml import etree
from ped import PlantExtract, Plant
from ssparser import SunSpecDataParser

# TODO get these Parser classes flatter and dumping non-Parser objects

xsd_filename = "sunspec_plant_extract.xsd"
xsd_dir = "xsd"
time_format = '%Y-%m-%dT%H:%M:%SZ'
#   logger = logging.getLogger('ped') f =
#   logging.Formatter('%(asctime)s:%(name)s:%(levelname)s:%(message)s') sh =
#   logging.StreamHandler() sh.setFormatter(f) logger.addHandler(sh)


class PlantExtractParser(object):
    """ A Plant Extract Document processor supporting straightforward
        interaction with information from each of the standard blocks.
    """

    element_name = "sunSpecPlantExtract"

    def __init__(self):
        logging.debug("PlantExtract.__init__()")
        self.xml = None

    def parse(self, ped_file):
        """Parse the plant extract document (minus sunSpecData)"""
        logging.debug("PlantExtract.parse()")
        # open file and validate
        this_dir, this_filename = os.path.split(__file__)
        xsd_file = os.path.join(this_dir, xsd_dir, xsd_filename)
        self.ped_file = ped_file
        self.tree = etree.parse(self.ped_file)
        schema_doc = etree.parse(xsd_file)
        self.schema = etree.XMLSchema(schema_doc)
        logging.info("PlantExtract.parse() valid_xml:" + str(self.valid_xml()))
        # now parse the Plant Extract
        self.envelope = self.tree.getroot()
        if self.envelope is None:
            raise PlantExtractException("sunSpecPlantExtract root not found.")

        logging.debug("PlantExtract.parse()")
        env = self.envelope
        time = env.get('t')
        if time is None:
            raise PlantExtractException("sunSpecPlantExtract root node 't' attribute is required.")
        else:
            # Plant Extract t="YYYY-MM-DDThh:mm:ssZ"
            self.time = dt.datetime.strptime(time, time_format)

        # check for sunSpecPlantExtract version, assume '2' if absent
        v = env.get('v')
        if v is not None:
            self.version = int(v)
        else:
            self.version = '2'

        # check for sunSpecPlantExtract seqId, assume '1' if absent
        sid = env.get('seqId')
        if sid is not None:
            int(sid)
            self.seqId = sid
        else:
            self.seqId = '1'

        # check for sunSpecPlantExtract lastSeqId, assume '1' if absent
        lid = env.get('lastSeqId')
        if lid is not None:
            int(lid)
            self.lastSeqId = lid
        else:
            self.lastSeqId = '1'

        if int(self.seqId) > int(self.lastSeqId):
            raise PlantExtractException("seqId can't be larger than the lastSeqId")

        self.plant = PlantParser().parse(self.envelope.find(PlantParser.element_name))
        self.sunspec_data = SunSpecDataParser().parse_block(
            self.envelope.find(SunSpecDataParser.element_name)
        )
        # TODO: sunSpecMetadata
        # TODO: strings
        # TODO: extract extensions
        return

    def parse_data(self):
        """Parse the sunSpecData block's Points"""
        if self.sunspec_data.exists and not self.sunspec_data.parsed:
            self.sunspec_data.parse_points()

    def __str__(self):
        """Produces a string representation of the Plant Extract envelope"""
        return ''.join(['PlantExtract v:', str(self.version), ' t:', str(self.time),
                        ' seqId:', str(self.seqId), ' lastSeqId:', str(self.lastSeqId)])

    def last(self):
        """Determines if this Plant Extract is the last extract in a set
        """
        return self.seqId == self.lastSeqId

    def valid_xml(self, assert_=False):
        """Determines if this Plant Extract is valid XML in compliance with the XSD"""
        if assert_:
            self.schema.assert_(self.tree)
            self.valid = True
        else:
            self.valid = self.schema.validate(self.tree)

        return self.valid

    def objectify(self):
        """ Creates and returns a non-parsing Python representation of the Plant
            Extract Document.
        """
        ped = PlantExtract(self.plant.objectify(), self.seqId, self.lastSeqId)
        ped.time = self.time
        return ped


    # @classmethod
    # def is_ped(cls, ped_file):
    #     """Convenience method to determine if the given file is a Plant Extract
    #     """
    #     is_ped_file = False
    #     try:
    #         ped = PlantExtract()
    #         ped.parse(ped_file)
    #         is_ped_file = True
    #     except PlantExtractException, pexe:
    #         is_ped_file = False
    #         logging.exception("PlantExtract.is_ped() PlantExtractException: %s", pexe)
    #     finally:
    #         return is_ped_file


class PlantParser(object):
    element_name = 'plant'

    def __init__(self):
        logging.debug("Plant.__init__()")

    def parse(self, element):
        self.element = pe = element
        if pe is None:
            raise PlantExtractException("plant element not found")

        plant_id = pe.get('id')
        if plant_id is None:
            raise PlantExtractException("plant 'id' attribute is required.")
        else:
            self.id = UUID(plant_id)

        v = pe.get('v')
        if v is not None:
            self.version = int(v)

        self.locale = pe.get('locale')
        self.name = get_node_value(pe, 'name')
        self.description = get_node_value(pe, 'description')
        self.notes = get_node_value(pe, 'notes')
        ad = get_node_value(pe, 'activationDate')
        if ad is not None:
            self.activation_date = dt.datetime.strptime(ad, '%Y-%m-%d')

        self.location = LocationParser(pe.find(LocationParser.element_name))
        self.name_plate = NamePlateParser(pe.find(NamePlateParser.element_name))
        self.capabilities = CapabilitiesParser(pe.find(CapabilitiesParser.element_name))
        self.participants = list()
        for participant in self.element.iter(Participant.element_name):
            self.participants.append(Participant(participant))

    def tostring(self):
        """Produces a string representation of some parsed Plant values """
        return ''.join(["Plant id:", self.id.hex, ", v:", str(self.version),
                        ", name:", self.name, ", locale:", self.locale,
                        ", description:", self.description])


class PropertyContainerParser(object):
    def __init__(self, my_element):
        self.properties = defaultdict(list)
        self.element = my_element
        if self.element is not None:
            for prop in self.element.iter(PropertyParser.element_name):
                prop_id = prop.get('id')
                prop_type = prop.get('type')
                self.properties[prop_id].append(PropertyParser(prop_id, prop_type,
                                                         prop.text))


class PropertyParser(object):
    element_name = 'property'

    def __init__(self, prop_id, prop_type, text):
        self.id = prop_id
        self.type = prop_type
        self.text = text


class LocationParser(PropertyContainerParser):
    element_name = 'location'

    def __init__(self, location_element):
        super(LocationParser, self).__init__(location_element)

    def parse(self):
        le = self.element
        self.latitude = get_node_value(le, 'latitude')
        self.longitude = get_node_value(le, 'longitude')
        self.line1 = get_node_value(le, 'line1')
        self.line2 = get_node_value(le, 'line2')
        self.city = get_node_value(le, 'city')
        self.stateProvince = get_node_value(le, 'stateProvince')
        self.country = get_node_value(le, 'country')
        self.postal = get_node_value(le, 'postal')
        self.elevation = get_node_value(le, 'elevation')
        self.timezone = get_node_value(le, 'timezone')


class NamePlateParser(PropertyContainerParser):
    element_name = 'namePlate'

    def __init__(self, nameplate_element):
        super(NamePlateParser, self).__init__(nameplate_element)


class CapabilitiesParser(PropertyContainerParser):
    element_name = 'capabilities'

    def __init__(self, capabilities_element):
        super(CapabilitiesParser, self).__init__(capabilities_element)
        print self.properties


class Participant(PropertyContainerParser):
    element_name = 'participant'

    def __init__(self, participant_element):
        super(Participant, self).__init__(participant_element)
        self.type = get_node_value(self.element, 'type')


class PlantExtractException(Exception):
    pass


#####################
#   Utility functions
def get_node_value(node, node_name):
    node_value = None

    if node is None:
        return None
    elif node.find(node_name) is not None:
        node_value = node.find(node_name).text

    return node_value

########################
#   command line parsing
if __name__ == '__main__':
    import argparse


    parser = argparse.ArgumentParser(description='Process or create Plant Extract Documents')
    parser.add_argument('--log', dest='loglevel', default='WARNING',
                        help='set the log level (default:WARNING)')
    parser.add_argument('--test', dest='activate_tests', action='store_true',
                        help='activate doctests for the Plant Extract class')
    sp = parser.add_subparsers()

    sp_create = sp.add_parser('create',
                              help="Create a plant extract document")

    sp_parse = sp.add_parser('parse', help="Parse a plant extract document")
    sp_parse.add_argument('--ped', type=file, nargs='+',
                          help='one or more plant extract documents to process - absolute path')
    sp_parse.add_argument('--xsd', type=file, nargs=1,
                          help='override the default XML schema document for validation')
    sp_parse.add_argument('--novalid', dest='validation', action='store_false',
                          help='do not validate the given plant extract documents')

    args = parser.parse_args()
    #   Now for some post parsing output
    # print args.ped

    if args.loglevel is not None:
        loglevel = args.loglevel.upper()
        print ">> Setting loglevel to: " + loglevel
        logging.getLogger().setLevel(loglevel)

    if args.activate_tests is True:
        """ Run some tests on the PlantExtract class
    >>>   ped = PlantExtract(args.ped[0])
    print ped.last()
    >>>   print 'TmpBOM'
    ps = ped.sunspec_data.get_matching_points('TmpBOM') for p in ps: print
      p.tostring()
    >>>   print 'TotWh'
    ps = ped.sunspec_data.get_matching_points('TotWh') for p in ps: print
    p.tostring() print "Points in period"
    #   start_time: 2012-10-28 22:00:59 end_time: 2012-10-28 22:03:00 > 5 or 6
    #   inclusive
    start_time = dt.datetime(2012, 10, 28, 22, 00, 00)
    end_time = dt.datetime(2012, 10, 28, 22, 03, 00)
    ped.sunspec_data.get_points_in_period(start_time, end_time, 'TotWh')
    """
    else:
        # xsd_full_file = os.path.join(os.getcwd(), xsd_dir, xsd_filename)
        ped = PlantExtractParser()
        ped.parse(args.ped[0])
        print ped
        print ">> PlantExtract Plant"
        print ped.plant
        print ">> PlantExtract parsing sunSpecData"
        ped.parse_data()
        print ">> PlantExtract completed parsing of sunSpecData"
        print ped.sunspec_data  # the sunSpecData block
        print ped.sunspec_data.device_records[0].models[0].smdx # SMDX info of model
        print ped.sunspec_data.device_records[0].models[0].points[0] # block structure




# Copyright 2013 Gridward LLC
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
import logging

from lxml import etree
from lxml import objectify


smdx_dir = "smdx"
smdx_prefix = "smdx"
smdx_ext = ".xml"
xsd_dir = "xsd"
smdx_xsd = "smdx.xsd"
smdx_schema_parser = None


class SMDX(object):
    # SMDX descriptive model id constants
    INVERTER_SINGLE_PHASE = 101
    INVERTER_SPLIT_PHASE = 102
    INVERTER_THREE_PHASE = 103
    METER_SINGLE_PHASE = 201

    _cache = {}  # holds the cached SMDX object if instantiated multiple times

    def __init__(self, element, model_id):
        logging.debug("SMDX.__init__()")
        self.points = {}
        self.smdx_tree = None
        return

    def __new__(cls, element, model_id):
        try:
            return SMDX._cache[model_id]
        except KeyError:
            logging.info("New SMDX object being instantiated for: " + model_id)
            # you must call __new__ on the base class
            x = super(SMDX, cls).__new__(cls)

            x.points = {}
            x.smdx_tree = None
            x.model_id = model_id
            smdx_filename = smdx_prefix + "_" + str.rjust(str(x.model_id), 5, "0") + smdx_ext
            print element, model_id, smdx_filename
            parser = get_smdx_parser()

            try:
                this_dir, this_filename = os.path.split(__file__)
                x.smdx_tree = objectify.parse(os.path.join(this_dir, smdx_dir, smdx_filename),
                                              parser)
            except IOError:
                logging.error("IOError caught while opening " + smdx_filename)

            if (x.smdx_tree is None):
                raise SunSpecSMDXException("SMDX object requires valid SMDX file: " + smdx_filename)

            x.__init__(element, model_id)
            SMDX._cache[model_id] = x
            logging.info("SMDX Cache: " + str(SMDX._cache))

            return x

    def get_points(self):
        root = self.smdx_tree.getroot()
        for b in root.model.block:
            for p in b.point:
                p_id = p.get('id')
                self.points[p_id] = SMDXPoint(p)

        return self.points

    def get_mandatory_points(self):
        man_points = {}
        points = self.get_points()
        for p in points.itervalues():
            if p.mandatory:
                man_points[p.id] = p

        return man_points


class SMDXPoint(object):
    """SMDX Point class that understands points like:
    <point id="PhVphA" offset="8" type="uint16" sf="V_SF" units="V" mandatory="true" />
    """

    # list of convenient point ID values
    ENERGY = 'WH'
    TOTAL_ENERGY_EXPORTED = 'TotWhExp'
    TOTAL_ENERGY_IMPORTED = 'TotWhImp'
    POWER = 'W'
    GLOBAL_HORIZONTAL_IRRADIANCE = 'GHI'
    PLANE_OF_ARRAY_IRRADIANCE = 'POAI'
    DIFFUSE_IRRADIANCE = 'DFI'

    FLOAT32 = 'float32'
    INT16 = 'int16'
    UINT16 = 'unint16'
    INT32 = 'int32'
    ACC16 = 'acc16'
    ACC32 = 'acc32'

    def __init__(self, element):
        element = element
        if (element is None) or (element.tag != 'point'):
            logging.warning("SMDXPoint objects must have a 'point' element")
            return

        self.id = element.get('id')
        logging.debug("SMDXPoint.__init__() self.id: " + self.id)
        self.len = element.get('len')
        self.offset = element.get('offset')
        self.type = element.get('type')
        self.sf = element.get('sf')
        self.units = element.get('units')
        self.access = element.get('access')
        self.value = element.text
        self.mandatory = False
        mand = element.get('mandatory')
        if mand is not None:
            self.mandatory = True if (mand.lower() == "true") else False

        logging.debug("SMDXPoint.__init__() self.mandatory: " + str(self.mandatory))


class SunSpecSMDXException(Exception):
    pass


def get_smdx_parser():
    global smdx_schema_parser
    if (smdx_schema_parser is None):
        schema_filename = os.path.join(os.getcwd(), xsd_dir, smdx_xsd)
        try:
            schema_etree = etree.parse(schema_filename)
            schema = etree.XMLSchema(schema_etree)
            logging.info("get_smdx_parser() getting a global lxml.objectify SMDX parser")
            smdx_schema_parser = objectify.makeparser(schema=schema)
        except IOError:
            logging.error("IOError caught while opening " + schema_filename)

    return smdx_schema_parser



    #     try:
    #       smdx_etree = etree.parse(os.path.join(os.getcwd(), smdx_dir, smdx_filename))
    #     except IOError:
    #       logging.warning("IOError caught while opening " + smdx_filename)
    #
    #     try:
    #       schema_etree = etree.parse(os.path.join(os.getcwd(), smdx_dir, smdx_xsd))
    #       schema = etree.XMLSchema(schema_etree)
    #     except IOError:
    #       logging.warning("IOError caught while opening " + smdx_xsd)
    #
    #     if (schema_etree is None) or (smdx_etree is None):
    #       return
    #     else:
    #       self.exists = True
    #       self.__valid(schema, smdx_etree)
    #
    #     smdx = SMDX(smdx_etree.getroot())
    #
    #
    #   def __valid(self, schema, smdx, assert_=False):
    #     """Determines if this SunSpecModel is valid XML in compliance with the SMDX XSD
    #     """
    #     if assert_:
    #       schema.assert_(smdx)
    #       self.valid = True
    #     else:
    #       self.valid = schema.validate(smdx)
    #       return self.valid
    #
    #   def parse(self):
    #
    #     return
    #
    # class SMDX(object):
    #   def __init__(self, root):
    #     self.root = root
    #     self.exists = False
    #     if self.root is None: return
    #     else:
    #       self.exists = True
    #       self.__parse()
    #       return
    #
    #   def __parse(self):
    #     """Parse the SMDX document
    #     """
    #     l = self.root.xpath('strings/model/label')
    #     self.label = l[0].text
    #
    #     self.points = defaultdict(list)
    #     point_elements = self.root.xpath('model//point')
    #     for p in point_elements:
    #       p_id = p.get('id')
    #       self.points[p_id].append(SMDXPoint(p))
    #
    #
    # class SMDXPoint(object):
    #   def __init__(self, element):
    #     e = element
    #     # check for sunSpecPlantExtract seqId, assume '1' if absent
    #     self.id     = e.get('id')
    #     self.offset = e.get('offset')
    #     self.type   = e.get('type')
    #     self.sf = e.get
    #
    #     return


# coding=utf-8
# Copyright 2013 Gridward LLC
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from uuid import UUID
from uuid import uuid4
from collections import defaultdict
import logging
import datetime as dt
import json

from sunspec import SunSpecData

# TODO move all parsing out into PlantParser, LocationParser, etc... which will
# TODO clean up the core objects.

time_format = '%Y-%m-%dT%H:%M:%SZ'
logger = logging.getLogger('ped')
f = logging.Formatter('%(asctime)s:%(name)s:%(levelname)s:%(message)s')
sh = logging.StreamHandler()
sh.setFormatter(f)
logger.addHandler(sh)


class PlantExtract(object):
    """ A Plant Extract Document
    """
    def __init__(self, plant, seqId='1', lastSeqId='1', ssd=None):
        """Create a plant extract document given parameters and data for the
           standard blocks.
        :param plant: a Plant object with at least a PlantID
        :param ssd: a sunspec_data.DeviceRecord[]
        """
        logger.debug("PlantExtract.create()")
        self.version = '2'
        self.seqId = seqId
        self.lastSeqId = lastSeqId
        self.time = dt.datetime.utcnow().strftime(time_format)
        self.plant = plant

    def __str__(self):
        """Produces a string representation of the Plant Extract envelope"""
        return ''.join(['PlantExtract v:', str(self.version), ' t:', str(self.time),
                        ' seqId:', str(self.seqId), ' lastSeqId:', str(self.lastSeqId)])

    def last(self):
        """Determines if this Plant Extract is the last extract in a set
        """
        return self.seqId == self.lastSeqId

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=2)


class Plant(object):
    def __init__(self, plant_id, locale="en-US", name="", notes="",
               description="", activation_date=None, location=None,
               name_plate=None, design_elements=None, array=None,
               equipment=None):
        """

        :param plant_id: a UUID4 that uniquely identifies the plant, required
        :param locale:
        :param name:
        :param notes:
        :param description:
        :param activation_date: the date the power plant was actively
            delivering power to the utility grid or energy off-taker. The
            activation_date value must be compliant with the ISO 8601 date
            format: "YYYY-MM-DD”
        :param location:
        :param name_plate:
        :param design_elements:
        :param array: a list of elements that describe the attributes of a PV
            array
        :param equipment:
        :return:
        """
        self.version = 2
        self.plant_id = plant_id
        self.locale = locale
        self.name = name
        self.notes = notes
        self.description = description
        if activation_date is not None:
            # validate given activation date complies with format
            ad = dt.datetime.strptime(activation_date, '%Y-%m-%d')
            # convert back into string
            self.activation_date = ad.strftime('%Y-%m-%d')

        self.location = location
        self.name_plate = name_plate
        self.design_elements = design_elements
        self.array = array
        self.equipment = equipment

    def __str__(self):
        """Produces a string representation of some parsed Plant values """
        return ''.join(["Plant id:", self.plant_id, ", v:", str(self.version),
                        ", name:", self.name, ", locale:", self.locale,
                        ", description:", self.description])

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=2)


class PropertyContainer(object):
    def __init__(self, props):
        #self.properties = defaultdict(list)
        self.properties = {}
        for p in props:
            self.properties[p.prop_id] = p

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=2)

    def __str__(self):
        pc = ''
        for k in self.properties:
            pc += str(self.properties[k]) + ' '
        return pc

class Property(object):
    def __init__(self, prop_id, prop_type, text):
        self.prop_id = prop_id
        self.type = prop_type
        self.text = text

    def __str__(self):
        return ''.join(['Property prop_id:', str(self.prop_id), ', type:',
                        str(self.type), ', text:', str(self.text)])


class Location(PropertyContainer):
    def __init__(self, latitude=0.0, longitude=0.0, line1="", line2="", city="",
               state_province="", country="", postal="", elevation="",
               timezone="", properties=defaultdict(list)):
        """ Create a Location object using the given information.
        :param latitude: the decimal degrees north or south of the equator of
            the plant that is its parent. The latitude value is
            a positive (representing northward) or negative (representing
            southward) decimal number in the expected format: ±XX.XXXXXX
            where the digits after the decimal are only necessary when
            expressing increasingly accurate values.
        :param longitude: the decimal degrees east or west from the Prime
            Meridian of the plant that is its parent. The longitude value
            is a positive (representing eastward) or negative (representing
            westward) decimal number in the expected format: ±XXX.XXXXXX
            where the digits after the decimal are necessary when expressing
            increasingly accurate values.
        :param line1:
        :param line2:
        :param city:
        :param state_province:
        :param country:
        :param postal:
        :param elevation:
        :param timezone:
        :param properties:
        :return location: the created Location object
        """

        super(Location, self).__init__(properties)

        self.latitude = str(latitude)
        self.longitude = str(longitude)
        self.line1 = line1
        self.line2 = line2
        self.city = city
        self.stateProvince = state_province
        self.country = country
        self.postal = postal
        self.elevation = elevation
        self.timezone = timezone
        self.properties = properties

    def __str__(self):
        return ''.join(["Location latitude:", self.latitude, ", longitude:",
                        str(self.longitude), ", line1:", str(self.line1),
                        ", line2:", str(self.line2), ", city:", str(self.city),
                        ", stateProvince:", str(self.stateProvince), ", country:",
                        str(self.country)])


class NamePlate(PropertyContainer):
    pass

class DesignElements(PropertyContainer):
    pass

class Array(PropertyContainer):
    def __init__(self, props, array_id=1):
        super(Array, self).__init__(props)
        self.array_id=array_id

class Equipment(PropertyContainer):
    def __init__(self, props, equipment_type='meter'):
        super(Equipment, self).__init__(props)
        self.equipment_type=equipment_type

class Capabilities(PropertyContainer):
    pass

class Participant(PropertyContainer):
    def __init__(self, props, participant_type="owner"):
        super(Participant, self).__init__(props)
        self.participant_type = participant_type


class PlantExtractException(Exception):
    pass


########################
#   command line parsing
if __name__ == '__main__':
    import argparse


    parser = argparse.ArgumentParser(description='Create a Plant Extract Document')
    parser.add_argument('--log', dest='loglevel', default='WARNING',
                        help='set the log level [default:WARNING]')
    parser.add_argument('--test', dest='activate_tests', action='store_true',
                        help='activate doctests for the Plant Extract class')
    # sp = parser.add_subparsers()
    #
    # sp_create = sp.add_parser('create',
    #                           help="Create a plant extract document")

    args = parser.parse_args()
    #   Now for some post parsing output
    # print args.ped

    if args.loglevel is not None:
        loglevel = args.loglevel.upper()
        print ">> Setting loglevel to: " + loglevel
        logger.setLevel(loglevel)

    if args.activate_tests is True:
        """ Run some tests on the PlantExtract class
    >>>   ped = PlantExtract(args.ped[0])
    print ped.last()
    >>>   print 'TmpBOM'
    ps = ped.sunspec_data.get_matching_points('TmpBOM') for p in ps: print
      p.tostring()
    >>>   print 'TotWh'
    ps = ped.sunspec_data.get_matching_points('TotWh') for p in ps: print
    p.tostring() print "Points in period"
    #   start_time: 2012-10-28 22:00:59 end_time: 2012-10-28 22:03:00 > 5 or 6
    #   inclusive
    start_time = dt.datetime(2012, 10, 28, 22, 00, 00)
    end_time = dt.datetime(2012, 10, 28, 22, 03, 00)
    ped.sunspec_data.get_points_in_period(start_time, end_time, 'TotWh')
    """
    else:
        ped = PlantExtract(
            Plant(
                uuid4().urn,
                activation_date="2013-03-02",
                location=Location(latitude=1.1, longitude=2.2,
                                         city="Redwood City",
                                         state_province="CA"),
                name_plate=NamePlate(props=[
                    Property('installedDCCapacity', 'float', '6.5'),
                    Property('installedACCapacity', 'float', '6.4')
                ]),
                design_elements=DesignElements(props=[
                    Property('plantType', 'string', 'commercial')
                ]),
                array=Array(props=[
                    Property('description','string','Carport')
                ], array_id=1),
                equipment=Equipment(props=[
                    Property('Mn', 'string', 'MeterManuf'),
                    Property('Md', 'string', 'MeterModel'),
                    Property('uncertainty', 'float', '0.5')
                ], equipment_type='meter')
            )
        )

        print ped
        print ped.plant
        print ped.plant.location
        print ped.plant.name_plate
        print ped.plant.equipment

        print "JSON PED", ped.toJSON()

        # ped.sunspec_data =

        # print etree.tostring(ped.xml, pretty_print=True)
        # ped = PlantExtract()
        # ped.parse(args.ped[0])
        # print ped
        # print ">> PlantExtract Plant"
        # print ped.plant
        # print ">> PlantExtract parsing sunSpecData"
        # ped.parse_points()
        # print ">> PlantExtract completed parsing of sunSpecData"
        # print ped.sunspec_data  # the sunSpecData block
        # print ped.sunspec_data.device_records[0].models[0].smdx # SMDX info of model
        # print ped.sunspec_data.device_records[0].models[0].points[0] # block structure




# Copyright 2013 Gridward LLC
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
from distutils.core import setup
from setuptools import find_packages

version = '0.2.9'
README = os.path.join(os.path.dirname(__file__), 'README.md')
long_description = open(README).read()

setup(name='plantextract',
      version=version,
      description='Python Plant Extract Document processor',
      long_description = long_description,
      author='Brett Francis',
      author_email='brett@gridward.com',
      url='https://github.com/lashex/ped.git',
      packages=['plantextract'],
      # packages=find_packages(),
      include_package_data=True,
      license='Apache License, Version 2.0',
      install_requires=['lxml == 3.2.1'],
      keywords = "sunspec plant extract",
      classifiers=['Development Status :: 2 - Pre-Alpha',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Other Audience',
                   'Natural Language :: English',
                   'Programming Language :: Python',
                   'Topic :: Utilities',
                   'Topic :: Software Development',
                   'Topic :: Text Processing :: Markup :: XML',
                   'Operating System :: OS Independent',],
      )


# Copyright 2013 Gridward LLC
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from operator import attrgetter
import datetime
import hashlib
import os
import logging

from smdx import SMDX
from smdx import SMDXPoint as SP


# logger = logging.getLogger('ped.ssd')


class SunSpecDataParser(object):
    element_name = 'sunSpecData'

    def __init__(self):
        logging.debug("SunSpecData.__init__()")

    def parse_block(self, element):
        """Parse the SunSpecData block device records, minus Points
        """
        if element is None:
            return

        self.element = element

        # check for sunSpecData version, assume '1' if absent
        v = self.element.get('v')
        if v is not None:
            self.version = int(v)
        else:
            self.version = 1

        self.device_records = list()
        for d_record in self.element.iter(DeviceRecordParser.element_name):
            self.device_records.append(DeviceRecordParser(d_record))

        self.parsed = False

    def parse_points(self):
        """Parse the SunSpecData block's Points
        """
        logging.debug("SunSpecData.parse()")
        for dr in self.device_records:
            dr.parse()
        self.parsed = True

    def get_points(self, model_id):
        all_points = list()
        for dr in self.device_records:
            for m in dr.models:
                if m.id == model_id:
                    all_points = all_points + m.points
        return all_points

    def _get_matching_points(self, model_id, point_id):
        """Get a list of points which match the given point_id
        """
        points = list()
        logging.debug(''.join(["SunSpecData.get_matching_points() looking for point_id:",
                               str(point_id)]))
        for dr in self.device_records:
            models = dr.models
            if models is not None:
                for m in models:
                    if m.id == model_id:
                        plist = m.get_matching_points(point_id)
                        if (plist is not None) and plist:
                            points = points + plist
        return points

    def get_points_in_period(self, start_time, end_time, model_id, point_id='All'):
        """Get a Point.time sorted list of points that are between the start_time
           and end_time and that match the point_id.

           If both start_time and end_time are None, then points for the sunSpecData
           block's entire time range is returned.

           Arguments:
           start_time -- the datetime describing the beginning of the period
           end_time -- the datetime describing the end of the period
           point_id -- the id of the Points to retrieve within the period

           Return:
           Points within the period as time sorted Points in a list
        """
        logging.info("SunSpecData.get_points_in_period: " + str(start_time) +
                     " > " + str(end_time) + " point_id: " + point_id)

        if point_id is 'All':
            points = self.get_points(model_id)
        else:
            points = self._get_matching_points(model_id, point_id)

        period_points = list()
        for p in points:
            if end_time is None and start_time is None:
                period_points.append(p)
            elif start_time < p.time < end_time:
                period_points.append(p)
        logging.info('Period points: ' + str(period_points))
        period_points = sorted(period_points, key=attrgetter('time'))
        return period_points

    def __str__(self):
        return ''.join(['SunSpecData v:', str(self.version),
                        ' device record count: ', str(len(self.device_records))])


class DeviceRecordParser(object):
    element_name = 'd'

    def __init__(self, element):
        element = element
        if element is None:
            logging.warning("DeviceRecord.__init__() element provided is None")
            return

        self.manufacturer = get_attr(element, 'man')
        self.model = get_attr(element, 'mod')
        self.serial_number = get_attr(element, 'sn')
        self.logger_id = get_attr(element, 'lid')
        self.logger_id_namespace = get_attr(element, 'ns')
        self.device_interface_id = get_attr(element, 'if')
        self.device_id = get_attr(element, 'id')
        self.correlation_id = get_attr(element, 'cid')

        self._determine_id()
        time = get_attr(element, 't')
        if time is None:
            raise SunSpecDataException("DeviceRecord 't' attribute is required.")
        else:
            # SunSpecData <d ... t="YYYY-MM-DDThh:mm:ssZ"
            self.time = datetime.datetime.strptime(time, '%Y-%m-%dT%H:%M:%SZ')

        self.models = list()
        for m in element.iter(ModelParser.element_name):
            m_id = m.get('id')
            if m_id is not None:
                logging.debug("DeviceRecord.__init__() adding model_id: " + m_id)
                mp = ModelParser()
                mp.parse(m, m_id, time)
                self.models.append(mp)

    def _determine_id(self):
        self.device_id_type = None
        gs = None
        if self.model and self.manufacturer and self.serial_number:
            gs = "" + self.serial_number + self.model + self.manufacturer
            self.device_id_type = 'Common Block'
        elif self.logger_id and self.device_id:
            gs = "" + self.device_id + self.logger_id
            self.device_id_type = 'Logger-specific, user-assigned'

        if gs is None:
            s = "A Device must have a globally unique identifier, either man: mod: sn: or lid: id:"
            raise SunSpecDataException(s)

        self.guid = int(hashlib.md5(gs).hexdigest(), 16)
        return

    def parse(self):
        logging.debug("SunSpecData.DeviceRecord.parse()")
        for model in self.models:
            model.parse()

    def tostring(self):
        s = "Device" + os.linesep
        s += " id_type:'" + self.device_id_type + "' guid:" + str(self.guid) + os.linesep
        s += " man:" + self.manufacturer + " mod:" + self.model + " sn:" + self.serial_number + os.linesep
        s += " lid:" + self.logger_id + " ns:" + self.logger_id_namespace
        s += " id:" + self.device_id + os.linesep
        s += " if:" + self.device_interface_id + " cid:" + self.correlation_id

        return s


class ModelParser(object):
    element_name = 'm'

    def parse(self, element, model_id, dr_time):
        self.dr_time = dr_time
        self.points = list()
        if element is None:
            raise SunSpecDataException("SunSpec model element is required for a Model")
        else:
            self.element = element

        if model_id is None:
            raise SunSpecDataException("SunSpec model element must have an id attribute")
        else:
            self.model_id = int(model_id)

        self.smdx = SMDX(element, model_id)
        logging.debug('Model constructed for model_id:' + str(model_id))

    def parse_data(self):
        """ Parse Points contained by the Model
        """
        logging.debug("Model.parse() model_id:" + str(self.model_id))

        logging.info("Model.parse() instantiating Points")
        for p in self.element.iter(PointParser.element_name):
            self.points.append(PointParser(self, p))

        smdx_points = self.smdx.get_points()
        logging.info("Model.parse() adding units from SMDXPoints to Points")
        for p in self.points:
            if p.id in smdx_points:
                logging.info("Point.id:%s found in SMDX.model_id: %s".format(
                    str(p.id), str(self.model_id))
                )
                p.unit = smdx_points[p.id].units
                p.type = smdx_points[p.id].type
            else:
                logging.warning(''.join(
                    ["Point.id:", str(p.id),
                     " exists but shouldn't for SMDX.model_id: ",
                     str(self.model_id)])
                )

        mand_met = False
        if self._has_mandatory_points() is None:
            mand_met = True
        logging.debug("Model.parse() has all mandatory points? " + str(mand_met))

    def get_matching_points(self, point_id):
        """Get a list of points which match the given point_id
        """
        points = list()
        logging.info(''.join(["Model.get_matching_points() looking for point_id:",
                              str(point_id)]))
        for p in self.points:
            logging.debug(''.join(["Model.get_matching_points() p.id:", p.id,
                                   " given point_id:", str(point_id)]))
            if p.id == point_id:
                points.append(p)

        return points

    def _has_mandatory_points(self):
        man_points = self.smdx.get_mandatory_points()
        points = self.points
        logging.debug("Model._has_mandatory_points()")
        missing_mand = {}

        for sp in man_points.itervalues():
            if sp.mandatory:
                found = False
                for p in points:
                    if sp.id == p.id:
                        logging.info("Found mandatory point:" + str(p.id))
                        found = True
                        break

                if not found:
                    logging.warning("Model.id:" + str(self.model_id) +
                                    " missing mandatory SMDXPoint.id:" + sp.id)
                    missing_mand[sp.id] = sp

        return missing_mand

    def tostring(self):
        return ''.join(['Model.id:', str(self.model_id), ' dr_time:', str(self.dr_time)])


class PointParser(object):
    element_name = 'p'

    def __init__(self, model, element):
        self.model = model
        self.id = element.get('id')
        self.x = element.get('x')
        self.value = element.text
        self.units = None
        self._type = None

        time = element.get('t')
        if time is None:
            self.time = model.dr_time
        else:
            # SunSpecData <p ... t="YYYY-MM-DDThh:mm:ssZ"
            self.time = datetime.datetime.strptime(time, '%Y-%m-%dT%H:%M:%SZ')

    def get_type(self):
        return self._type

    def set_type(self, type_value):
        """

        :param type_value:
        :return:
        """
        self._type = type_value
        print self.id
        logging.info("Point.set_type() id:", self.id, " type:", self._type)
        if (self._type == SP.UINT16 or self._type == SP.INT16
            or self._type == SP.INT32 or self._type == SP.ACC16
            or self._type == SP.ACC32):
            self.value = int(self.value)  # Convert into 32 bit signed
        elif self._type == SP.FLOAT32:
            self.value = float(self.value)   # Convert into 64 bit float
        return

    # noinspection PyShadowingBuiltins
    type = property(get_type, set_type)

    #     def __radd__(self, other):
    #         result = None
    #         if (self.type == SP.UINT16) or (self.type == SP.INT16) or (self.type == SP.INT32):
    #             print self.tostring()
    #             if (other is not None):
    #                 result = other + int(self.value)
    #             else:
    #                 result = int(self.value)
    #         elif self.type == SP.FLOAT32:
    #             if (other is not None):
    #                 return other + float(self.value)
    #             else:
    #                 return float(self.value)
    #         return result
    #
    #     def __add__(self, other):
    #         if self.type == SP.UINT16 or SP.INT16 or SP.INT32:
    #             result = int(self.value) + other
    #         elif self.type == SP.FLOAT32:
    #             result = float(self.value) + other
    #         return result

    def tostring(self):
        return ''.join(['Point.id:', self.id, ' x:', str(self.x), ' value:',
                        str(self.value), ' units:', str(self.units), ' type:',
                        str(self.type), ' time:', str(self.time)])


class SunSpecDataException(Exception):
    def __init__(self, argument):
        super(SunSpecDataException, self).__init__(argument)
        self.argument = argument


####################
# Utility functions
def get_attr(element, attr_name):
    attr_value = None
    if element is None:
        return attr_value
    elif element.get(attr_name):
        attr_value = element.get(attr_name)
    return attr_value
